Unit GameStat;

Interface

Uses Keyboard,G320,IntvDbg,AVI;

Type
  TRGB = Packed Record
    R,G,B: Byte;
  End; { TRGB }
  PPalette  = ^TIPalette;
  TIPalette = Array[0..255] Of TRGB;
  TPalette  = TIPalette;
  TSaveRec  = Record
    Unknown1   : Byte;                   { 20000 }
    PSG2_2     : Array[0..6] Of Byte;    { 20001-20007 }
    PSG1       : Array[0..$D] Of Byte;   { 20008-20015 }
    Unknown2   : Byte;                   { 20016 }
    PSG2_1     : Array[0..6] Of Byte;    { 20017-2001D }
    Regs       : CP1600Regs;             { 2001E-2002D }
    Unknown3   : Byte;                   { 2002E }
    StatusWord : Byte;                   { 2002F }       { Not sure about this }
    Unknown4   : Word;                   { 20030-20031 } { Always holds 0001h }
    Unknown5   : Word;                   { 20032-20033 }
    Unknown6   : Byte;                   { 20034 }       { Always holds 00h }
    IntrLoc    : Word;                   { 20035-20036 } { Always holds 1004h }
    Unknown7   : Word;                   { 20037-20038 } { Always holds 0175h }
    FB_CS      : Byte;                   { 20039 }
    Unknown8   : Byte;                   { 2003A }       { Always holds 01h }
  End;

Const
  BI_RGB  = 0;

  BGColor = 18;
  C1Color = 16;
  C2Color = 20;
  TxColor = 19;
  RedText = $1213;
  BlkText = $1200;

  _Palette: Array[0..22] Of TRGB =
   ((R:$00; G:$00; B:$00),(R:$00; G:$00; B:$37),(R:$3A; G:$00; B:$00),(R:$30; G:$34; B:$1D),
    (R:$01; G:$21; B:$00),(R:$00; G:$2C; B:$00),(R:$3F; G:$38; B:$06),(R:$3F; G:$3F; B:$3F),
    (R:$2A; G:$2A; B:$2A),(R:$15; G:$2F; B:$33),(R:$3F; G:$23; B:$03),(R:$0D; G:$18; B:$0A),
    (R:$3F; G:$15; B:$15),(R:$26; G:$27; B:$3F),(R:$15; G:$34; B:$00),(R:$32; G:$05; B:$14),
    (R:$3F; G:$3F; B:$3F),(R:$38; G:$38; B:$38),(R:$32; G:$32; B:$32),(R:$2F; G:$00; B:$00),
    (R:$1F; G:$20; B:$3F),(R:$17; G:$18; B:$3F),(R:$10; G:$10; B:$3F));
Var
  Palette   : TIPalette;

Procedure GameStatus;
Procedure ScreenCapture(ExtraLines: Boolean);
Function  EnterStringPopup(Title,Prompt,OldStr: String): String;
Function  CreateFile(Var F: File; St: String): Boolean;
Function  CanOpenFile(St: String): Boolean;
Procedure SetPalette(Var Pal: TIPalette; Dim: Boolean);
Procedure LoadGame(Buf: CP1600Mem; IOWriteProc,IOReadProc: TIOProc;
                   Var Regs: CP1600Regs; StatusWord: Word);
Procedure QuickLoad(Buf: CP1600Mem; IOWriteProc,IOReadProc: TIOProc;
                   Var Regs: CP1600Regs; StatusWord: Word;
                   FileName : String);
Procedure QuickSave(Buf: CP1600Mem; IOWriteProc,IOReadProc: TIOProc;
                   Var Regs: CP1600Regs; StatusWord: Word; FB: Boolean;
                   FileName : String);
Function  ImportBasicPgm(Var FileName: String): Boolean;
Function  ImportECSPgm(Var FileName: String): Boolean;
Function  FileDialog(Const Title,Filter: String): String;

Implementation

Uses Time,IntvMnu,SmalFont,VGAGUI;

Const
  _32BIT_OPERAND = $66;

Var
  OldKey    : Array[0..255] of Byte;
  Saved     : Integer;
  I         : Integer;

Procedure SaveKeys;
Begin
  If Saved = 0 Then
  Begin
    Move(Key,OldKey,SizeOf(OldKey));
    ExitKeyboard;
  End;
  Inc(Saved);
End; { SaveKeys }

Procedure RestoreKeys;
Begin
  If Saved > 0 Then
  Begin
    Dec(Saved);
    If Saved = 0 Then
    Begin
      InitKeyboard;
      Move(OldKey,Key,SizeOf(Key));
    End;
  End;
End; { RestoreKeys }

Procedure SetPalette(Var Pal: TIPalette; Dim: Boolean);
Var I: Integer;
Begin
  Port[$3C8] := 0;
  For I := 0 To 255 Do
  Begin
    If (I < 16) And Dim Then
    Begin
      Port[$3C9] := Pal[I].R Shr 3;
      Port[$3C9] := Pal[I].G Shr 3;
      Port[$3C9] := Pal[I].B Shr 3;
    End
    Else
    Begin
      Port[$3C9] := Pal[I].R Shr 2;
      Port[$3C9] := Pal[I].G Shr 2;
      Port[$3C9] := Pal[I].B Shr 2;
    End;
  End; { For I }
End; { SetPalette }

Procedure SaveScreen(Buf: Pointer);
Begin
  Asm
    CLD
    PUSH  DS
    MOV   CX,160*200
    LES   DI,DWORD PTR Buf
    MOV   AX,0A000h
    MOV   DS,AX
    SUB   SI,SI
@L1:
    MOV   AL,BYTE PTR [SI]
    MOV   BYTE PTR ES:[DI],AL
    ADD   SI,2
    INC   DI
    DEC   CX
    JNZ   @L1
    POP   DS
  End; { Asm }
End; { SaveScreen }


Procedure RestoreScreen(Buf: Pointer);
Begin
  Asm
    CLD
    PUSH  DS
    MOV   CX,160*200
    MOV   AX,0A000h
    MOV   ES,AX
    SUB   DI,DI
    LDS   SI,DWORD PTR Buf
@L1:
    MOV   AL,BYTE PTR [SI]
    MOV   AH,AL
    MOV   WORD PTR ES:[DI],AX
    INC   SI
    ADD   DI,2
    DEC   CX
    JNZ   @L1
    POP   DS
  End; { Asm }
End; { RestoreScreen }

Procedure Dim(BufSize: Word); Assembler;
Asm
  MOV   AX,0A000h
  MOV   ES,AX
  MOV   CX,WORD PTR BufSize
  SHR   CX,1
  SUB   DI,DI
@L1:
  MOV   DX,WORD PTR [ES:DI]
  MOV   BX,DX
  AND   BX,0E0Eh
  AND   DX,0F0F0h

  OR    BL,BL
  JZ    @L2
  SHR   BL,2
  CMP   BL,2
  JGE   @L2
  MOV   BL,2
@L2:
  OR    BH,BH
  JZ    @L3
  SHR   BH,2
  CMP   BH,2
  JGE   @L3
  MOV   BH,2
@L3:
  OR    DX,BX
  MOV   WORD PTR [ES:DI],DX
  ADD   DI,2
  LOOP  @L1
End; { Dim }

Procedure DrawWindow(X1,Y1,X2,Y2: Word; BK,C1,C2: Byte);
Const BorderWidth = 5;
Var
  X,Y   : Word;
  Addr1 : Word;
  Addr2 : Word;
  CInc  : Integer;

Begin

  { Draw top and bottom borders }

  For Y := 0 To BorderWidth - 1 Do
  Begin
    Addr1 := 320 * (Y1 + Y);
    Addr2 := 320 * (Y2 + Y - BorderWidth + 1);
    CInc  := (Abs((BorderWidth Div 2) - Integer(Y)));
    For X := X1 + 1 To X2 - 1 Do
    Begin
      If ((X + Y + Y1) And 2) <> 0
       Then Mem[$A000:Addr1 + X] := C1 + CInc
       Else Mem[$A000:Addr1 + X] := C2 + CInc;
      If ((X + Y + Y2 - BorderWidth + 1) And 2) <> 0
       Then Mem[$A000:Addr2 + X] := C1 + CInc
       Else Mem[$A000:Addr2 + X] := C2 + CInc;
    End; { For X }
  End; { For Y }

  { Draw left and right borders }

  For Y := Y1 + 1 To Y2 - 1 Do
  Begin
    Addr1 := 320 * Y;
    For X := X1 To X1 + BorderWidth - 1 Do
    Begin
      CInc  := (Abs((BorderWidth Div 2) - Integer(X - X1)));
      If ((X + Y) And 2) <> 0
       Then Mem[$A000:Addr1 + X] := C1 + CInc
       Else Mem[$A000:Addr1 + X] := C2 + CInc;
    End; { For X }
    For X := X2 - BorderWidth + 1 To X2 Do
    Begin
      CInc  := (Abs((BorderWidth Div 2) - Integer(X - (X2 - BorderWidth + 1))));
      If ((X + Y) And 2) <> 0
       Then Mem[$A000:Addr1 + X] := C1 + CInc
       Else Mem[$A000:Addr1 + X] := C2 + CInc;
    End; { For X }
  End; { For Y }

  { Draw center area }

  For Y := Y1 + BorderWidth To Y2 - BorderWidth Do
   FillChar(Mem[$A000:320 * Y + X1 + BorderWidth],X2 - X1 - 2 * BorderWidth + 1,BK);
End; { DrawWindow }

Procedure PutChar(X,Y: Word; C: Char; Color: Word); Assembler;
Asm
  MOV   AX,WORD PTR Y

  SUB   BX,BX
  MOV   BH,AL
  SHL   AX,6
  ADD   BX,AX
  MOV   DI,BX
{
  MOV   BX,320
  MUL   BX
  MOV   DI,AX
}
  ADD   DI,WORD PTR X
  PUSH  DS
  LDS   SI,DWORD PTR FontPtr
  MOV   AL,BYTE PTR C
  SUB   AH,AH
  SHL   AX,3
  ADD   SI,AX
  MOV   AX,0A000h
  MOV   ES,AX
  MOV   AL,8
  ADD   DI,7
  MOV   DX,WORD PTR Color
@Loop:
  MOV   BH,BYTE PTR [SI]
  MOV   CX,8
  MOV   BL,1
@L1:
  TEST  BH,BL
  JZ    @L2
  MOV   BYTE PTR [ES:DI],DL
  DEC   DI
  ADD   BL,BL
  LOOP  @L1
  ADD   DI,320
  INC   SI
  ADD   DI,8
  DEC   AL
  JNZ   @Loop
  POP   DS
  JMP   @L3
@L2:
  MOV   BYTE PTR [ES:DI],DH
  DEC   DI
  ADD   BL,BL
  LOOP  @L1
  ADD   DI,320
  INC   SI
  ADD   DI,8
  DEC   AL
  JNZ   @Loop
  POP   DS
@L3:
End; { PutChar }

Procedure WriteOut(X,Y: Word; St: String; Color: Word);
Var I: Byte;
Begin
  If St = '' Then Exit;
  For I := 0 To Length(St) - 1 Do PutChar(X + (I * 8),Y,St[I + 1],Color);
End; { WriteOut }

Procedure GameStatus;
Var
  Buf     : Pointer;

  I : Integer;

  Procedure DrawDiff(X,Y: Word; Diff: Boolean; Width: Word);
  Begin
    If Diff Then WriteOut(X,Y,'A',$089F) Else WriteOut(X,Y,'B',$089F);
  End; { DrawDiff }

  Procedure DrawYN(X,Y: Word; Diff: Boolean; Width: Word);
  Begin
    If Diff Then WriteOut(X,Y,'Y',$089F) Else WriteOut(X,Y,'N',$089F);
  End; { DrawDiff }

Begin
  Key[kEsc] := 0;
  SetPalette(Palette,True);
  InitGUI(320,200,False,True);
  DrawWindow(100,15,220,185,BGColor,C1Color,C2Color);
(*
  For I := 0 To 255 Do
  Begin
    Mem[$A000:I]       := I;
    Mem[$A000:I + 320] := I;
    Mem[$A000:I + 640] := I;
    Mem[$A000:I + 960] := I;
  End; { For I }
*)
  WriteOut(108,24,'INTVDOS 1.0b',RedText);
  WriteSmOut(108,40, 'Esc - Exit game',BlkText);
  WriteSmOut(108,50, 'F1  - Help',BlkText);
  WriteSmOut(108,60, 'F2  - Quick save',BlkText);
  WriteSmOut(108,80, 'F4  - Quick load',BlkText);
  WriteSmOut(108,110,'F7  - Import pgm',BlkText);
  WriteSmOut(108,120,'F8  - Load game',BlkText);
{
  WriteSmOut(108,120,'F8  - Save info',BlkText);
}
  WriteSmOut(108,130,'F9  - Sc.capture',BlkText);
  WriteSmOut(108,140,'F10 - Reset',BlkText);
  WriteSmOut(108,150,'F11 - Save to AVI',BlkText);
  WriteSmOut(108,160,'F12 - Toggle spd',BlkText);
{
  WriteSmOut(108,150,'F11 - Save game',BlkText);
  WriteSmOut(108,160,'F12 - Load game',BlkText);
}
  Repeat
  Until Key[kEsc] <> 0;
  Repeat
  Until Key[kEsc] = 0;
  SetPalette(Palette,False);
End; { GameStatus }


Function EnterStringPopup(Title,Prompt,OldStr: String): String;
Var
  Buf     : Pointer;
  BufSize : LongInt;
  F       : File;
  App     : PVGAApplication;

Begin
  SaveKeys;
  BufSize := 160 * 200;
  GetMem(Buf,BufSize);
  SaveScreen(Buf);
  App := New(PVGAApplication,Init(320,200,False,False,True));
  EnterStringPopup := Trim(App^.InputBox(Title,Prompt,OldStr));
  Dispose(App,Done);
  RestoreScreen(Buf);
  FreeMem(Buf,BufSize);
  RestoreKeys;
End; { EnterStringPopup }


Function CanOpenFile(St: String): Boolean;
Var
  Buf     : Pointer;
  BufSize : LongInt;
  Error   : Boolean;
  App     : PVGAApplication;

Begin
  SaveKeys;
  Error   := False;
  BufSize := 160 * 200;
  St      := Trim(St);
  If (St = '') Or Not Exist(St) Then Error := True;
  If Error Then
  Begin
    Repeat
    Until Key[kEsc] = 0;
    If St <> '' Then
    Begin
      GetMem(Buf,BufSize);
      SaveScreen(Buf);
      App := New(PVGAApplication,Init(320,200,False,False,True));
      App^.ShowMessage('Error','Could not open file',[mbOK]);
      Dispose(App,Done);
      RestoreScreen(Buf);
      FreeMem(Buf,BufSize);
    End;
    CanOpenFile := False;
  End
  Else CanOpenFile := True;
  RestoreKeys;
End; { CanOpenFile }


Function CreateFile(Var F: File; St: String): Boolean;
Var
  Buf      : Pointer;
  BufSize  : LongInt;
  Error    : Boolean;
  Continue : Boolean;
  App     : PVGAApplication;

Begin
  SaveKeys;
  Error   := False;
  BufSize := 160 * 200;
  If St = '' Then Error := True;

  Continue := True;
  If (Not Error) And Exist(St) Then
  Begin
    Key[kY] := 0;
    Key[kN] := 0;
    GetMem(Buf,BufSize);
    SaveScreen(Buf);
    App := New(PVGAApplication,Init(320,200,False,False,True));
    Continue := (App^.ShowMessage('Warning: file already exists','Overwrite file ' + St +
                              '?',[mbOK,mbCancel]) = idOK);
    Dispose(App,Done);
    RestoreScreen(Buf);
    FreeMem(Buf,BufSize);
  End;
  If Error Then Continue := False;

  If Continue Then
  Begin
    Assign(F,St);
    {$I-}
    ReWrite(F,1);
    {$I+}
    If IOResult <> 0 Then Error := True;
    If Error Then
    Begin
      Repeat
      Until Key[kEsc] = 0;
      If St <> '' Then
      Begin
        GetMem(Buf,BufSize);
        SaveScreen(Buf);
        App := New(PVGAApplication,Init(320,200,False,False,True));
        App^.ShowMessage('Error','Could not create file',[mbOK]);
        Dispose(App,Done);
        RestoreScreen(Buf);
        FreeMem(Buf,BufSize);
      End;
      CreateFile := False;
    End
    Else CreateFile := True;
  End
  Else CreateFile := False;
  RestoreKeys;
End; { CreateFile }


Procedure ScreenCapture(ExtraLines: Boolean);
Var
  Buf     : Pointer;
  BufSize : LongInt;
  St      : String;
  F       : File;
  App     : PVGAApplication;

  Procedure MakeBMPFile(Var F: File; ExtraLines: Boolean);
  Var
    BF      : BITMAPFILEHEADER;
    BI      : BITMAPINFOHEADER;
    I       : Integer;
    Q       : RGBQuad;

  Begin
    FillChar(BF,SizeOf(BF),#0);
    FillChar(BI,SizeOf(BI),#0);

    BF.bfType[0] := 'B';
    BF.bfType[1] := 'M';
    BF.bfSize    := SizeOf(BF) + SizeOf(BI) + 256 * SizeOf(RGBQuad) + BufSize;
    BF.bfOffBits := SizeOf(BF) + SizeOf(BI) + 256 * SizeOf(RGBQuad);

    BI.biSize          := 40;
    BI.biWidth         := 320;
    BI.biHeight        := 200;  { Top-down bitmap }
    BI.biPlanes        := 1;
    BI.biBitCount      := 8;
    BI.biCompression   := BI_RGB;
    BI.biSizeImage     := 0;
    BI.biXPelsPerMeter := 2000;
    BI.biYPelsPerMeter := 2000;
    BI.biClrUsed       := 256;
    BI.biClrImportant  := 256;

    BlockWrite(F,BF,SizeOf(BF));
    BlockWrite(F,BI,SizeOf(BI));

    For I := 0 To 255 Do
    Begin
      Q.R := Palette[I].B Shl 2;
      Q.G := Palette[I].G Shl 2;
      Q.B := Palette[I].R Shl 2;
      Q.Reserved := 0;
      BlockWrite(F,Q,SizeOf(Q));
    End; { For I }
    For I := 0 To 199 Do
     If ExtraLines
      Then BlockWrite(F,Mem[$A000:(201 - I) * 320],320)
      Else BlockWrite(F,Mem[$A000:(199 - I) * 320],320);
  End; { MakeBMPFile }

Begin
  SaveKeys;
  SetPalette(Palette,True);
  App := New(PVGAApplication,Init(320,200,False,False,True));
  St  := UpperCase(Trim(App^.FileDialog('Save screenshot to BMP','*.*','')));
  Dispose(App,Done);
  If CreateFile(F,St) Then
  Begin
    MakeBMPFile(F,ExtraLines);
    Close(F);
  End;
  SetPalette(Palette,False);
  RestoreKeys;
End; { ScreenCapture }

Function ImportBasicPgm(Var FileName: String): Boolean;
Begin
  Key[kEsc] := 0;
  SetPalette(Palette,True);
  FileName := FileDialog('Import ECS Basic text program','*.*');
  ImportBasicPgm := CanOpenFile(FileName);
  SetPalette(Palette,False);
End; { ImportBasicPgm }

Function ImportECSPgm(Var FileName: String): Boolean;
Begin
  Key[kEsc] := 0;
  SetPalette(Palette,True);
  FileName := FileDialog('Load ECS program from cassette','*.*');
  ImportECSPgm := CanOpenFile(FileName);
  SetPalette(Palette,False);
End; { ImportECSPgm }

Function FileDialog(Const Title,Filter: String): String;
Var
  Buf     : Pointer;
  BufSize : LongInt;
  St      : String;
  App     : PVGAApplication;

Begin
  SaveKeys;
  BufSize := 160 * 200;
  GetMem(Buf,BufSize);
  SaveScreen(Buf);
  App := New(PVGAApplication,Init(320,200,False,False,True));
  St  := UpperCase(Trim(App^.FileDialog(Title,Filter,'')));
  Dispose(App,Done);
  RestoreScreen(Buf);
  FreeMem(Buf,BufSize);
  RestoreKeys;
  FileDialog := St;
End; { FileDialog }

Procedure LoadGame(Buf: CP1600Mem; IOWriteProc,IOReadProc: TIOProc;
                   Var Regs: CP1600Regs; StatusWord: Word);
Var
  St  : String;
  I   : Integer;
  F   : File;
  B   : Byte;
  FB  : Byte;
  Rec : TSaveRec;

Begin
  Key[kEsc] := 0;
  SetPalette(Palette,True);
  St := FileDialog('Load game state','*.*');
  QuickLoad(Buf,IOWriteProc,IOReadProc,Regs,StatusWord,St);
  SetPalette(Palette,False);
End; { LoadGame }

Procedure QuickLoad(Buf: CP1600Mem; IOWriteProc,IOReadProc: TIOProc;
                   Var Regs: CP1600Regs; StatusWord: Word;
                   FileName : String);
Var
  St  : String;
  I   : Integer;
  F   : File;
  B   : Byte;
  FB  : Byte;
  Rec : TSaveRec;

Begin
  Key[kEsc] := 0;
  If (FileName <> '') And CanOpenFile(FileName) Then
  Begin
    Assign(F,FileName);
    Reset(F,1);

    { Load CP1600 memory }

    For I := 0 To $F Do BlockRead(F,Buf[I]^[0],$2000);
    BlockRead(F,Rec,SizeOf(Rec));
    Close(F);

    { Load CP1600 status word }

    StatusWord := Rec.StatusWord;

    { Load CP1600 registers }

    Regs := Rec.Regs;

    { Load sound information }

    For I := 0 To $F Do IOWriteProc($1F0 + I,Rec.PSG1[I]);
    For I := 0 To 6  Do IOWriteProc($F0  + I,Rec.PSG2_1[I]);
    For I := 0 To 6  Do IOWriteProc($F7  + I,Rec.PSG2_1[I]);

    { Load foreground/background / color stack information }

    For I := 0 To $65 Do IOWriteProc(I,Buf[0]^[I]);
    If Rec.FB_CS <> 0 Then IOWriteProc($0021,0) Else IOReadProc($0021,0);
  End;
End; { QuickLoad }

Procedure QuickSave(Buf: CP1600Mem; IOWriteProc,IOReadProc: TIOProc;
                   Var Regs: CP1600Regs; StatusWord: Word; FB: Boolean;
                   FileName : String);
Var
  St  : String;
  I   : Integer;
  F   : File;
  B   : Byte;
  Rec : TSaveRec;

Begin
  Key[kEsc] := 0;
  If FileName <> '' Then
  Begin
    Assign(F,FileName);
    ReWrite(F,1);

    { Save CP1600 memory }

    For I := 0 To $F Do BlockWrite(F,Buf[I]^[0],$2000);

    { Initialize save record }

    FillChar(Rec,SizeOf(Rec),#0);

    { Save CP1600 status word }

    Rec.StatusWord := StatusWord;

    { Save CP1600 registers }

    Rec.Regs := Regs;

    { Save sound information }

    For I := 0 To $F Do Rec.PSG1[I]   := Buf[0]^[$1F0 + I];
    For I := 0 To 6  Do Rec.PSG2_1[I] := Buf[0]^[$F0  + I];
    For I := 0 To 6  Do Rec.PSG2_2[I] := Buf[0]^[$F7  + I];

    { Save foreground/background / color stack information }

    If FB Then Rec.FB_CS := 1 Else Rec.FB_CS := 0;

    { Save other information }


    Rec.Unknown4 := $0001;
    Rec.Unknown5 := $FFFF;
    Rec.Unknown6 := $00;
    Rec.IntrLoc  := $1004;
    Rec.Unknown7 := $0175;
    Rec.Unknown8 := $01;

    BlockWrite(F,Rec,SizeOf(Rec));
    Close(F);
  End;
End; { QuickSave }

Begin
  Saved := 0;
  FillChar(Palette,SizeOf(Palette),#0);
  For I := 0 To 22 Do
  Begin
    Palette[I].R := _Palette[I].R Shl 2;
    Palette[I].G := _Palette[I].G Shl 2;
    Palette[I].B := _Palette[I].B Shl 2;
  End; { For I }
{  Move(_Palette,Palette,SizeOf(_Palette));}
End.
