Unit INTVMnu;
Interface

Uses ProtMenu,TypeText,IntvCntr;

Type
  TColorScheme   = (csGameType,csAlphabetical,csEvery3,csEvery4,csEvery5,csEvery6);
  TDisplayScheme = (dsNameExt4,dsNameExt3,dsNameExt2,
                    dsNameOnly4,dsNameOnly3,dsNameOnly2,
                    dsRealNameLeft,dsRealNameCenter);
  ConfigRec     = Record
    DoSound            : Boolean;
    ScreenModeWidth    : Word;
    ScreenModeHeight   : Word;
    ScreenModeDepth    : Word;
    MenuColorScheme    : TColorScheme;
    MenuForeColor1     : Byte;
    MenuForeColor2     : Byte;
    MenuBackColor1     : Byte;
    MenuBackColor2     : Byte;
    MenuDisplayScheme  : TDisplayScheme;
    FrameRate          : Word;
    PlaybackRate       : Word;
    Exec               : String;
    ECS                : String;
    GROM               : String;
    IVoiceROM          : String;
    ExtraLines         : Boolean;
  End;
  TCartName = String[12];
  TCart = (Cart5,Cart4,CartECS);
  PFileType = ^TFileType;
  TFileType = Record
    Name     : PChar; {TCartName;}
    Cart     : TCart;
    Next     : PFileType;
    RealName : PChar; {TRealCartName;}
    MemStr   : PChar; {TMemStr;}
  End;

Var
  Cart       : TCart;
  ProfileRec : PFileType;

Procedure InitConfig(Var Config: ConfigRec; Playback_Freq: Word;
                     Version: String);

Procedure SaveConfiguration(Var Config: ConfigRec; Version: String);

Procedure ErrorFileNotFound(Const Path,St: String);

Function Menu(Version: String;
              Var Top,Sel: Integer;
              Var Playback_Freq: Word;
              Var DoSound,UseMenu: Boolean;
              Var Config: ConfigRec): String;

Function IntToStr(L: LongInt): String;

Function Exist(FileName: String): Boolean;

Function Trim(St: String): String;

Function UpperCase(St: String): String;

Function ValHex(St: String): LongInt;

Implementation

Uses Crt,Dos,HexWrite,Time,INI,Keyboard,IntvDbg,Strings;

Const
  ConfigFile     = 'INTVDOS.INI';
  CRLF           = #13 + #10;
  _32BIT_OPERAND = $66;
  OutFileName    = 'INTVDOS.PRO';
  CartID : Array[TCart] Of String[2] = ('C5','C4','EC');
  DisplayColWidth: Array[TDisplayScheme] Of Byte = (16,15,14,12,11,10,40,40);

Type
  PIndex = ^TIndex;
  TIndex = Array[0..255] Of PFileType;

Var
  FHead       : PFileType;
  FTail       : PFileType;
  Buf         : Array[0..255] Of Byte;
  OldExitProc : Pointer;
  FileIndex   : PIndex;

Function ValHex(St: String): LongInt;
Var
  Initial : LongInt;
  I       : Integer;
  Value   : LongInt;

Begin
  If St = '' Then
  Begin
    ValHex := -1;
    Exit;
  End;
  Value   := 0;
  St      := UpperCase(St);
  I       := (Length(St) - 1) * 4;
  Asm
    DB    _32BIT_OPERAND
    SUB   AX,AX
    MOV   AX,1
    MOV   CX,WORD PTR I
    DB    _32BIT_OPERAND
    SHL   AX,CL
    DB    _32BIT_OPERAND
    MOV   WORD PTR Initial,AX
  End; { Asm }
  For I := 1 To Length(St) Do
  Begin
    If St[I] In ['0'..'9'] Then Inc(Value,Initial * (Ord(St[I]) - Ord('0')))
     Else
      If St[I] In ['A'..'F'] Then Inc(Value,Initial * (Ord(St[I]) - Ord('A') + 10))
       Else
       Begin
         ValHex := -1;
         Exit;
       End;
    Initial := Initial Shr 4;
  End; { For I }
  ValHex := Value;
End; { ValHex }

Function IntToStr(L: LongInt): String;
Var St: String;
Begin
  Str(L,St);
  IntToStr := St;
End; { IntToStr }

Function Trim(St: String): String;
Begin
  While (St <> '') And (St[1]          In [#9,' ']) Do St := Copy(St,2,Length(St) - 1);
  While (St <> '') And (St[Length(St)] In [#9,' ']) Do St := Copy(St,1,Length(St) - 1);
  Trim := St;
End; { Trim }

Function UpperCase(St: String): String;
Var I: Integer;
Begin
  For I := 1 To Length(St) Do St[I] := UpCase(St[I]);
  UpperCase := St;
End; { UpperCase }

Function Exist(FileName: String): Boolean;
Var Fil: File;
Begin
  Assign(Fil,FileName);
  {$I-}
  Reset(Fil);
  {$I+}
  If (IOResult = 0) Then
  Begin
    Close(Fil);
    Exist := True;
  End
  Else Exist := False;
End; { Exist }

Procedure SortFileList;
Var
  F1,F2 : PFileType;
  Cart  : TCart;
  P     : PChar;
  St    : String;

Begin
  F1 := FHead;
  While F1^.Next <> Nil Do
  Begin
    St := StrPas(F1^.Name);
    F2 := F1^.Next;
    While F2 <> Nil Do
    Begin
      If StrPas(F2^.Name) < St Then
      Begin
        P            := F1^.Name;
        F1^.Name     := F2^.Name;
        F2^.Name     := P;
        Cart         := F1^.Cart;
        F1^.Cart     := F2^.Cart;
        F2^.Cart     := Cart;
        P            := F1^.RealName;
        F1^.RealName := F2^.RealName;
        F2^.RealName := P;
        P            := F1^.MemStr;
        F1^.MemStr   := F2^.MemStr;
        F2^.MemStr   := P;
        St           := StrPas(F1^.Name);
      End;
      F2 := F2^.Next;
    End; { While }
    F1 := F1^.Next;
  End;
End; { SortFileList }

Procedure LoadFileList(Path: String);
Var S: SearchRec;
Begin
  FindFirst(Path + '*.BIN',Archive,S);
  While DOSError = 0 Do
  Begin
    If FHead = Nil Then
    Begin
      New(FHead);
      FTail := FHead;
    End
    Else
    Begin
      New(FTail^.Next);
      FTail := FTail^.Next;
    End;
    FTail^.Name     := StrPCopyNew(UpperCase(S.Name));
    FTail^.Cart     := Cart5;
    FTail^.Next     := Nil;
    FTail^.RealName := StrPCopyNew('');
    FTail^.MemStr   := StrPCopyNew('');
    FindNext(S);
  End; { While }
End; { LoadFileList }

Procedure Setup;
Begin
  FHead := Nil;
  FTail := Nil;
End; { Setup }

Procedure Cleanup;
Var FW1,FW2: PFileType;
Begin
  FW1 := FHead;
  While FW1 <> Nil Do
  Begin
    FW2 := FW1^.Next;
    Dispose(FW1);
    FW1 := FW2;
  End; { While }
  FHead := Nil;
  FTail := Nil;
  FillChar(FileIndex^,SizeOf(FileIndex^),#0);
End; { Cleanup }

Procedure OutputFile(Path: String);
Var
  F       : PFileType;
  Cart    : TCart;
  B       : Byte;
  I       : Integer;
  Done    : Boolean;
  St      : String;
  C       : Char;
  Lines   : Integer;
  OutFile : Text;

Begin
  C := #0;
  Assign(OutFile,OutFileName);
  ReWrite(OutFile);
  WriteLn(OutFile,'; INTVDOS cartridge profile list');
  WriteLn(OutFile,'; Valid cart codes: C5 C4 EC');
  WriteLn(OutFile);
  F     := FHead;
  Lines := 0;
  While F <> Nil Do
  Begin
    St := StrPas(F^.Name);
    If St[1] <> C Then
    Begin
      C := St[1];
      Screen.Write(C + ' ');
    End;
    Cart := Cart5;
    While Length(St) < 15 Do St := St + ' ';
    I  := 3 * Integer(F^.Cart);
    While I > 0 Do
    Begin
      St := St + ' ';
      Dec(I);
    End; { While }
    St := St + CartID[F^.Cart] + ' "' + StrPas(F^.RealName) + '"';
    If StrPas(F^.MemStr) <> '' Then
    Begin
      While Length(St) < 60 Do St := St + ' ';
      St := St + 'MEM=' + StrPas(F^.MemStr);
    End;
    WriteLn(OutFile,St);
    F := F^.Next;
    Inc(Lines);
    If (Lines Mod 10) = 0 Then
     WriteLn(OutFile,'; ---------------------------------------------------');
  End; { While }
  Close(OutFile);
  Screen.WriteLn('');
End; { OutputFile }

Procedure IndexFileList;
Var
  FWork : PFileType;
  C     : Char;
  St    : String;

Begin
  FillChar(FileIndex^,SizeOf(FileIndex^),#0);
  FWork := FHead;
  C     := '*';   { Choose a character that cannot be in the filename }
  While FWork <> Nil Do
  Begin
    St := StrPas(FWork^.Name);
    If St <> '' Then
    Begin
      If St[1] <> C Then
      Begin
        C                  := St[1];
        FileIndex^[Ord(C)] := FWork;
      End;
    End;
    FWork := FWork^.Next;
  End; { While }
End; { IndexFileList }

Procedure MakeProfile(Path: String);
Begin
  Cleanup;
  Setup;
  LoadFileList(Path);
  SortFileList;
  IndexFileList;
  OutputFile(Path);
End; { MakeProfile }

Procedure LoadProfile;
Var
  St,St1  : String;
  I       : Integer;
  F       : PFileType;
  Found   : Boolean;
  Cart    : TCart;
  OutFile : Text;

Begin
  If Exist(OutFileName) Then
  Begin
    Assign(OutFile,OutFileName);
    Reset(OutFile);
    While Not Eof(OutFile) Do
    Begin
      ReadLn(OutFile,St);
      St := Trim(St);
      If (St <> '') And (St[1] <> ';') Then
      Begin
        I  := Pos(' ',St);
        If I <> 0 Then
        Begin
          St1   := Trim(Copy(St,1,I - 1));
          St    := Trim(Copy(St,I + 1,Length(St) - I));
          If FHead = Nil Then
          Begin
            New(FHead);
            FTail := FHead;
          End
          Else
          Begin
            New(FTail^.Next);
            FTail := FTail^.Next;
          End;
          FTail^.Name     := StrPCopyNew(UpperCase(St1));
          FTail^.Next     := Nil;
          FTail^.Cart     := Cart5;
          FTail^.RealName := StrPCopyNew('');
          FTail^.MemStr   := StrPCopyNew('');
          I := Pos(' ',St);

          If I <> 0 Then
          Begin
            St1 := Trim(Copy(St,1,I - 1));
            St  := Trim(Copy(St,I + 1,Length(St) - I));
          End
          Else St1 := '';

          For Cart := Cart5 To CartECS Do
           If CartID[Cart] = St1 Then FTail^.Cart := Cart;

          { Locate the game's real name }

          I := Pos('"',St);
          If I <> 0 Then
          Begin
            St := Copy(St,I + 1,Length(St) - I);
            I  := Pos('"',St);
            If I <> 0 Then
            Begin
              St1 := Trim(Copy(St,1,I - 1));
              St  := Trim(Copy(St,I + 1,Length(St) - I));
              If St1 <> '' Then
              Begin
                StrDispose(FTail^.RealName);
                FTail^.RealName := StrPCopyNew(St1);
              End;
            End;
          End;

          { Load the memory placement information }

          I := Pos('MEM=',UpperCase(St));
          If I <> 0 Then
          Begin
            St := Trim(Copy(St,I + 4,Length(St) - I - 3));
            I  := Pos(' ',St);
            If I <> 0 Then
            Begin
              St1 := Trim(Copy(St,1,I - 1));
              St  := Trim(Copy(St,I + 1,Length(St) - I));
              If St1 <> '' Then FTail^.MemStr := StrPCopyNew(St1);
            End
            Else
            Begin
              If St <> '' Then FTail^.MemStr := StrPCopyNew(St);
            End;
          End;
        End;
      End;
    End; { While }
    Close(OutFile);
  End;
End; { LoadProfile }

{$F+}
Procedure KillProfile;
Begin
  Cleanup;
  Dispose(FileIndex);
  ExitProc := OldExitProc;
End; { KillProfile }
{$F-}

Function FindProfile(Name: TCartName; Var Prof: PFileType): TCart;
Var
  FWork : PFileType;
  Found : Boolean;
  St    : String;

Begin
  Name  := UpperCase(Name);
  Found := False;
  If Name <> '' Then
  Begin
    FWork := FileIndex^[Ord(Name[1])];
    If FWork <> Nil Then St := StrPas(FWork^.Name);
    While (FWork <> Nil) And (St[1] = Name[1]) And Not Found Do
    Begin
      If Name = St Then Found := True Else
      Begin
        FWork := FWork^.Next;
        If FWork <> Nil Then St := StrPas(FWork^.Name);
      End;
    End; { While }
  End;
  If Found Then
  Begin
    FindProfile := FWork^.Cart;
    Prof        := FWork;
  End
  Else
  Begin
    FindProfile := Cart5;
    Prof        := Nil;
  End;
End; { FindProfile }

Procedure InitConfig(Var Config: ConfigRec; Playback_Freq: Word;
                     Version: String);
Var
  F  : PINIFile;
  St : String;
  B  : Byte;

Begin
  Config.DoSound            := True;
  Config.FrameRate          := 60;
  Config.PlaybackRate       := Playback_Freq;
  Config.ScreenModeWidth    := 80;
  Config.ScreenModeHeight   := 25;
  Config.ScreenModeDepth    := 1;
  Config.MenuColorScheme    := csGameType;
  Config.MenuForeColor1     := LightGray;
  Config.MenuForeColor2     := White;
  Config.MenuBackColor1     := Blue;
  Config.MenuBackColor2     := Blue;
  Config.MenuDisplayScheme  := dsNameExt4;

  { Initalize controller interfaces }

  InterfaceHandler^.AddInterface(New(PMouseInterface,Init));
  InterfaceHandler^.AddInterface(New(PKeyboardInterface,Init));
  InterfaceHandler^.AddInterface(New(PJoystickInterface,Init));
  InterfaceHandler^.AddInterface(New(PGrIPInterface,Init));
  InterfaceHandler^.AddInterface(New(PSideWinderInterface,Init));
  InterfaceHandler^.AddInterface(New(PExternalInterface,Init));
{  InterfaceHandler^.AddInterface(New(PLogitechInterface,Init));}

  { Load from the config file }

  If Exist(ConfigFile) Then
  Begin
    F := New(PINIFile,Init(ConfigFile));
    Config.DoSound           := F^.ReadBool('GENERAL','DOSOUND',False);
    Config.FrameRate         := F^.ReadInteger('GENERAL','FRAMERATE',60);
    Config.PlaybackRate      := F^.ReadInteger('GENERAL','PLAYBACKRATE',Playback_Freq);
    Config.ScreenModeWidth   := F^.ReadInteger('GENERAL','SCREENMODEWIDTH',80);
    Config.ScreenModeHeight  := F^.ReadInteger('GENERAL','SCREENMODEHEIGHT',25);
    Config.ScreenModeDepth   := F^.ReadInteger('GENERAL','SCREENMODEDEPTH',8);
    Config.MenuColorScheme   := TColorScheme(F^.ReadInteger('GENERAL','MENUCOLORSCHEME',LongInt(csGameType)));
    Config.MenuForeColor1    := F^.ReadInteger('GENERAL','MENUFORECOLOR1',LightGray);
    Config.MenuForeColor2    := F^.ReadInteger('GENERAL','MENUFORECOLOR2',White);
    Config.MenuBackColor1    := F^.ReadInteger('GENERAL','MENUBACKCOLOR1',Blue);
    Config.MenuBackColor2    := F^.ReadInteger('GENERAL','MENUBACKCOLOR2',Blue);
    Config.MenuDisplayScheme := TDisplayScheme(F^.ReadInteger('GENERAL','MENUDISPLAYSCHEME',LongInt(dsNameExt4)));
    Config.Exec              := F^.ReadString('GENERAL','EXEC','EXEC.BIN');
    Config.ECS               := F^.ReadString('GENERAL','ECS','ECS.BIN');
    Config.GROM              := F^.ReadString('GENERAL','GROM','GROM.BIN');
    Config.IVoiceROM         := F^.ReadString('GENERAL','IVOICEROM','IVOICE.BIN');
    Config.ExtraLines        := F^.ReadBool('GENERAL','EXTRALINES',False);

    { Load controller interface information }

    InterfaceHandler^.LoadConfiguration(F);

    Dispose(F,Done);
  End
  Else SaveConfiguration(Config,Version);
End; { InitConfig }

Procedure SaveConfiguration(Var Config: ConfigRec; Version: String);
Var
  F  : PINIFile;
  St : String;
  F1 : File;

Begin
  If Exist(ConfigFile) Then
  Begin
    Assign(F1,ConfigFile);
    Erase(F1);
  End;
  F := New(PINIFile,Init(ConfigFile));
  F^.WriteBool('GENERAL','DOSOUND',Config.DoSound);
  F^.WriteInteger('GENERAL','FRAMERATE',Config.FrameRate);
  F^.WriteInteger('GENERAL','PLAYBACKRATE',Config.PlaybackRate);
  F^.WriteInteger('GENERAL','SCREENMODEWIDTH',Config.ScreenModeWidth);
  F^.WriteInteger('GENERAL','SCREENMODEHEIGHT',Config.ScreenModeHeight);
  F^.WriteInteger('GENERAL','SCREENMODEDEPTH',Config.ScreenModeDepth);
  F^.WriteInteger('GENERAL','MENUCOLORSCHEME',LongInt(Config.MenuColorScheme));
  F^.WriteInteger('GENERAL','MENUFORECOLOR1',Config.MenuForeColor1);
  F^.WriteInteger('GENERAL','MENUFORECOLOR2',Config.MenuForeColor2);
  F^.WriteInteger('GENERAL','MENUBACKCOLOR1',Config.MenuBackColor1);
  F^.WriteInteger('GENERAL','MENUBACKCOLOR2',Config.MenuBackColor2);
  F^.WriteInteger('GENERAL','MENUDISPLAYSCHEME',LongInt(Config.MenuDisplayScheme));
  F^.WriteString('GENERAL','EXEC',Config.Exec);
  F^.WriteString('GENERAL','ECS',Config.ECS);
  F^.WriteString('GENERAL','GROM',Config.GROM);
  F^.WriteString('GENERAL','IVOICEROM',Config.IVoiceROM);
  F^.WriteBool('GENERAL','EXTRALINES',Config.ExtraLines);
  InterfaceHandler^.SaveConfiguration(F);
  Dispose(F,Done);
End; { SaveConfiguration }

Procedure ErrorFileNotFound(Const Path,St: String);
Begin
  WriteLn('Sorry, but INTVDOS was unable to find file ' + St + ' in');
  WriteLn('directory ' + FExpand(Path));
  WriteLn('');
  WriteLn('You should make sure that INTVDOS was run from the');
  WriteLn('right directory and that ' + St + ' is present.');
End; { ErrorFileNotFound }
(*
Procedure MainMenuHelp;
Begin
  Screen.TextColor(LightGray);
  Screen.TextBackground(Blue);
  Screen.ClrScr;
  Screen.GotoXY(5,1);  Screen.Write('<Esc> ...... Quit   <F10> ..... Reset');
  Screen.GotoXY(5,3);  Screen.Write('     PLAYER 1            PLAYER 2');
  Screen.GotoXY(5,4);  Screen.Write('  +---+---+---+       +---+---+---+  ');
  Screen.GotoXY(5,5);  Screen.Write('  |   |   |   |       |   |   |   |  ');
  Screen.GotoXY(5,6);  Screen.Write('  | 7 | 8 | 9 |       | 1 | 2 | 3 |  ');
  Screen.GotoXY(5,7);  Screen.Write('  |   |   |   |       |   |   |   |  ');
  Screen.GotoXY(5,8);  Screen.Write('  +---+---+---+       +---+---+---+  ');
  Screen.GotoXY(5,9);  Screen.Write('F6|   |   |   |F6   F1|   |   |   |F1');
  Screen.GotoXY(5,10); Screen.Write('  | 4 | 5 | 6 |       | Q | W | E |  ');
  Screen.GotoXY(5,11); Screen.Write('  |   |   |   |       |   |   |   |  ');
  Screen.GotoXY(5,12); Screen.Write('  +---+---+---+       +---+---+---+  ');
  Screen.GotoXY(5,13); Screen.Write('F7|   |   |   |F8   F2|   |   |   |F3');
  Screen.GotoXY(5,14); Screen.Write('  | 1 | 2 | 3 |       | A | S | D |  ');
  Screen.GotoXY(5,15); Screen.Write('  |   |   |   |       |   |   |   |  ');
  Screen.GotoXY(5,16); Screen.Write('  +---+---+---+       +---+---+---+  ');
  Screen.GotoXY(5,17); Screen.Write('  |   |   |   |       |   |   |   |  ');
  Screen.GotoXY(5,18); Screen.Write('  | 0 | . |   |       | Z | X | C |  ');
  Screen.GotoXY(5,19); Screen.Write('  |INS|DEL|ENT|       |   |   |   |  ');
  Screen.GotoXY(5,20); Screen.Write('  +---+---+---+       +---+---+---+  ');
  Screen.GotoXY(5,21); Screen.Write('                                     ');
  Screen.GotoXY(5,22); Screen.Write('       UP                   I        ');
  Screen.GotoXY(5,23); Screen.Write('  LEFT    RIGHT          J     L     ');
  Screen.GotoXY(5,24); Screen.Write('      DOWN                  K        ');
  ReadKey;
  Screen.ClrScr;
  Screen.TextColor(LightGray);
  Screen.TextBackground(Blue);
  Screen.ClrScr;
End; { MainMenuHelp }
*)

Function Menu(Version: String;
              Var Top,Sel: Integer;
              Var Playback_Freq: Word;
              Var DoSound,UseMenu: Boolean;
              Var Config: ConfigRec): String;
Const
  ButtonNum : Array[1..15] Of Byte = (1,2,2,3,3,3,3,4,4,4,4,4,4,4,4);
  ButtonBit : Array[1..15] Of Byte = (1,2,2,4,4,4,4,8,8,8,8,8,8,8,8);

Type
  PFileListType = ^TFileListType;
  TFileListType = Record
    Name : String[12];
    Size : LongInt;
    Next : PFileListType;
    Prev : PFileListType;
    Prof : PFileType;
  End;

Var
  I       : Word;
  Head    : PFileListType;
  Tail    : PFileListType;
  Work    : PFileListType;
  Work1   : PFileListType;
  Temp    : TFileListType;
  S       : SearchRec;
  Ch,C    : Char;
  Cols    : Integer;
  Rows    : Integer;
  Num     : Integer;
  Show    : Boolean;
  J       : Integer;
  JoyX1   : LongInt;
  JoyY1   : LongInt;
  JoyX2   : LongInt;
  JoyY2   : LongInt;
  CJoyX   : LongInt;
  CJoyY   : LongInt;
  JoyX1_2 : LongInt;
  JoyY1_2 : LongInt;
  JoyX2_2 : LongInt;
  JoyY2_2 : LongInt;
  JoyB    : Byte;
  F       : File;
  Path    : String[79];
  _Menu   : PMenu;
  PreMenu : PMenu;
  DisMenu : PMenu;
  ColMenu : PMenu;
  ScnMenu : PMenu;
  EnaMenu : PMenu;
  _Win    : PWindow;
  MWid    : Integer;
  St      : String;
  Code    : Integer;
  DJX,DJY : LongInt;
  OldTop  : Integer;
  ScnMode : PScreenMode;
  Intf    : PInterface;
  Cap     : TCapability;
  Info    : TIntvControls;
  OldWork : PFileListType;
  Found   : Boolean;
  ContDir : Byte;

  Procedure Display;
  Var
    I,J  : Integer;
    W    : PFileListType;
    St   : String;
    Row  : Integer;
    Col  : Integer;
    Show : Boolean;

    Procedure ShowPartial(Const St: String; Const X,Y: Integer);
    Var
      I   : Integer;
      St1 : String[1];

    Begin
      For I := 1 To Length(St) Do
      Begin
        St1 := St[I];
        If St1 <> ' ' Then Screen.FastWrite(X + I - 1,Y,Screen.TextAttr,St1);
      End; { For I }
    End; { ShowPartial }

  Begin
    I   := Rows * Cols;
    W   := Work1;
    Row := 1;
    Col := 1;
    If Cols * DisplayColWidth[Config.MenuDisplayScheme] < MaxX Then Inc(I,Rows);
    While (W <> Nil) And (I <> 0) Do
    Begin
      Show := (W = Work) Or (W = OldWork) Or (Top <> OldTop);
      If W = Work Then
      Begin
        Screen.TextColor(Black);
        Screen.TextBackground(LightGray);
      End
      Else
      Begin
        Case Config.MenuColorScheme Of
          csGameType:
          Begin
            Screen.TextColor(LightGray);
            Screen.TextBackground(Blue);
          End;
          csAlphabetical:
          Begin
            If (Ord(W^.Name[1]) And 1) <> 0 Then
            Begin
              Screen.TextColor(Config.MenuForeColor1);
              Screen.TextBackground(Config.MenuBackColor1);
            End
            Else
            Begin
              Screen.TextColor(Config.MenuForeColor2);
              Screen.TextBackground(Config.MenuBackColor2);
            End;
          End;
          csEvery3:
          Begin
            If (((Row - 1) Div 3) And 1) <> 0 Then
            Begin
              Screen.TextColor(Config.MenuForeColor1);
              Screen.TextBackground(Config.MenuBackColor1);
            End
            Else
            Begin
              Screen.TextColor(Config.MenuForeColor2);
              Screen.TextBackground(Config.MenuBackColor2);
            End;
          End;
          csEvery4:
          Begin
            If (((Row - 1) Div 4) And 1) <> 0 Then
            Begin
              Screen.TextColor(Config.MenuForeColor1);
              Screen.TextBackground(Config.MenuBackColor1);
            End
            Else
            Begin
              Screen.TextColor(Config.MenuForeColor2);
              Screen.TextBackground(Config.MenuBackColor2);
            End;
          End;
          csEvery5:
          Begin
            If (((Row - 1) Div 5) And 1) <> 0 Then
            Begin
              Screen.TextColor(Config.MenuForeColor1);
              Screen.TextBackground(Config.MenuBackColor1);
            End
            Else
            Begin
              Screen.TextColor(Config.MenuForeColor2);
              Screen.TextBackground(Config.MenuBackColor2);
            End;
          End;
          csEvery6:
          Begin
            If (((Row - 1) Div 6) And 1) <> 0 Then
            Begin
              Screen.TextColor(Config.MenuForeColor1);
              Screen.TextBackground(Config.MenuBackColor1);
            End
            Else
            Begin
              Screen.TextColor(Config.MenuForeColor2);
              Screen.TextBackground(Config.MenuBackColor2);
            End;
          End;
          Else
            Screen.TextColor(LightGray);
            Screen.TextBackground(Blue);
        End; { Case }
      End;
      If Show Then
      Begin
        Case Config.MenuDisplayScheme Of
        dsNameExt4,dsNameExt3,dsNameExt2:
        Begin
          St := W^.Name;
          J  := Pos('.',St);
          If J <> 0 Then
           While J < 9 Do
           Begin
             Insert(' ',St,J);
             Inc(J);
           End; { While }
        End;
        dsNameOnly4,dsNameOnly3,dsNameOnly2:
        Begin
          St := W^.Name;
          J  := Pos('.',St);
          If J <> 0 Then St := Copy(St,1,J - 1);
        End;
        dsRealNameLeft:
        Begin
          If (W^.Prof <> Nil) And (StrLen(W^.Prof^.RealName) > 0)
           Then St := StrPas(W^.Prof^.RealName)
           Else St := W^.Name;
        End;
        dsRealNameCenter:
        Begin
          If (W^.Prof <> Nil) And (StrLen(W^.Prof^.RealName) > 0)
           Then St := StrPas(W^.Prof^.RealName)
           Else St := W^.Name;
          While Length(St) < DisplayColWidth[Config.MenuDisplayScheme] Do
          Begin
            St := St + ' ';
            If Length(St) < DisplayColWidth[Config.MenuDisplayScheme] Then
             St := ' ' + St;
          End; { While }
        End;
        Else
          St := '';
        End; { Case }
        While Length(St) < DisplayColWidth[Config.MenuDisplayScheme] Do St := St + ' ';
        If Col >= Cols Then
        Begin
          If Col * DisplayColWidth[Config.MenuDisplayScheme] > MaxX Then
           St := Copy(St,1,MaxX - (Col - 1) * DisplayColWidth[Config.MenuDisplayScheme]);
        End;
        Screen.FastWrite((Col - 1) * DisplayColWidth[Config.MenuDisplayScheme] + 1,Row + 1,Screen.TextAttr,St);
      End;
      Inc(Row);
      If Row >= Rows + 1 Then
      Begin
        Row := 1;
        Inc(Col);
      End;
      Dec(I);
      W := W^.Next;
    End; { While }

    { Erase any remaining entries }

    If Top <> OldTop Then
    Begin
      St := '';
      While Length(St) < DisplayColWidth[Config.MenuDisplayScheme] Do St := St + ' ';
      Screen.TextColor(LightGray);
      Screen.TextBackground(Blue);
      While I > 0 Do
      Begin
        Case Config.MenuColorScheme Of
        csGameType:
        Begin
          Screen.TextColor(LightGray);
          Screen.TextBackground(Blue);
        End;
        csAlphabetical:
        Begin
          Screen.TextColor(Config.MenuForeColor1);
          Screen.TextBackground(Config.MenuBackColor1);
        End;
        csEvery3:
        Begin
          If (((Row - 1) Div 3) And 1) <> 0 Then
          Begin
            Screen.TextColor(Config.MenuForeColor1);
            Screen.TextBackground(Config.MenuBackColor1);
          End
          Else
          Begin
            Screen.TextColor(Config.MenuForeColor2);
            Screen.TextBackground(Config.MenuBackColor2);
          End;
        End;
        csEvery4:
        Begin
          If (((Row - 1) Div 4) And 1) <> 0 Then
          Begin
            Screen.TextColor(Config.MenuForeColor1);
            Screen.TextBackground(Config.MenuBackColor1);
          End
          Else
          Begin
            Screen.TextColor(Config.MenuForeColor2);
            Screen.TextBackground(Config.MenuBackColor2);
          End;
        End;
        csEvery5:
        Begin
          If (((Row - 1) Div 5) And 1) <> 0 Then
          Begin
            Screen.TextColor(Config.MenuForeColor1);
            Screen.TextBackground(Config.MenuBackColor1);
          End
          Else
          Begin
            Screen.TextColor(Config.MenuForeColor2);
            Screen.TextBackground(Config.MenuBackColor2);
          End;
        End;
        csEvery6:
        Begin
          If (((Row - 1) Div 6) And 1) <> 0 Then
          Begin
            Screen.TextColor(Config.MenuForeColor1);
            Screen.TextBackground(Config.MenuBackColor1);
          End
          Else
          Begin
            Screen.TextColor(Config.MenuForeColor2);
            Screen.TextBackground(Config.MenuBackColor2);
          End;
        End;
        Else
          Screen.TextColor(LightGray);
          Screen.TextBackground(Blue);
        End; { Case }
        If Col >= Cols Then
        Begin
          If Col * DisplayColWidth[Config.MenuDisplayScheme] > MaxX Then
           St := Copy(St,1,MaxX - (Col - 1) * DisplayColWidth[Config.MenuDisplayScheme]);
        End;
        Screen.FastWrite((Col - 1) * DisplayColWidth[Config.MenuDisplayScheme] + 1,Row + 1,Screen.TextAttr,St);
        Inc(Row);
        If Row >= Rows + 1 Then
        Begin
          Row := 1;
          Inc(Col);
        End;
        Dec(I);
      End; { While }
    End;

    { Display the top status line }

    Screen.TextColor(Magenta);
    Screen.TextBackground(LightGray);
    Screen.GotoXY(1,1);
    Screen.ClrEol;
    Screen.GotoXY(1,MaxY);
    Screen.ClrEol;

    St := 'INTVDOS ' + Version + '   ' + IntToStr(MemAvail);
    Screen.FastWrite(2,1,Screen.TextAttr,St);

    Screen.TextColor(Black);

    St := Work^.Name;
    Screen.FastWrite(22,1,Screen.TextAttr,St);
    Str(Work^.Size,St);
    Screen.FastWrite(40,1,Screen.TextAttr,St);

    Screen.TextColor(Red);

    St := 'F10       Esc     ';
    ShowPartial(St,52,1);
    Screen.TextColor(Black);
    St := '    Menu      Quit';
    ShowPartial(St,52,1);

{
    St := 'F1       F10       Esc     ';
    ShowPartial(St,52,1);
    Screen.TextColor(Black);
    St := '   Help      Menu      Quit';
    ShowPartial(St,52,1);
}
    { Display the bottom status line }

    Str(Num,St);
    St := St + ' Files';
    Screen.FastWrite(2,MaxY,Screen.TextAttr,St);
{      St := '000';
      Screen.FastWrite(18,MaxY,Screen.TextAttr,St);}

    St := 'CS';
    Screen.FastWrite(25,MaxY,Screen.TextAttr,St);

    Screen.TextColor(Red);

{      St := '4      E';}
    St := '       E';
    ShowPartial(St,17,MaxY);
{    End;}
    OldTop  := Top;
    OldWork := Work;
  End; { Display }

  Procedure FindTop;
  Var I: Integer;
  Begin
    I     := Top;
    Work1 := Head;
    While (Work1 <> Nil) And (I > 0) Do
    Begin
      Work1 := Work1^.Next;
      Dec(I);
    End; { While }
  End;

  Procedure FindSel;
  Var I: Integer;
  Begin
    I     := Sel;
    Work  := Head;
    While (Work <> Nil) And (I > 0) Do
    Begin
      Work := Work^.Next;
      Dec(I);
    End; { While }
  End; { FindSel }

  Procedure _Up;
  Begin
    If Sel > 0 Then
    Begin
      If Sel = Top Then
      Begin
        Dec(Top);
        Work1 := Work^.Prev;
      End;
      Dec(Sel);
      Work := Work^.Prev;
      Show := True;
    End;
  End; { _Up }

  Procedure _Down;
  Begin
    If Sel < Num - 1 Then
    Begin
      If Sel = Top + Cols * Rows - 1 Then
      Begin
        Inc(Top);
        Work1 := Work1^.Next;
      End;
      Inc(Sel);
      Work := Work^.Next;
      Show := True;
    End;
  End; { _Down }

  Procedure _Left;
  Begin
    If Sel >= Rows Then
    Begin
      Show := True;
      Dec(Sel,Rows);
      If Sel < Top Then Dec(Top,Rows);
      If Top < 0 Then Top := 0;
      FindTop;
      FindSel;
    End;
  End; { _Left }

  Procedure _Right;
  Begin
    If Sel < Num - Rows Then
    Begin
      Show := True;
      Inc(Sel,Rows);
      If Sel >= Top + Cols * Rows Then Inc(Top,Rows);
      FindTop;
      FindSel;
    End;
  End; { _Right }

  Procedure Popup(St: String);
  Var
    W       : DWindow;
    I,J,X,Y : Integer;

  Begin
    Screen.TextColor(Black);
    Screen.TextBackground(LightGray);
    I := Length(St);
    X := (MaxX Div 2) - ((I + 4) Div 2);
    Y := MaxY Div 2;
    W.Init(X,Y - 2,X + I + 3,Y + 2,
           Black,LightGray,Black,LightGray,Black,LightGray,
           DoubleBorder,'','',True);
    W.Display;
    Screen.GotoXY(2,2);
    Screen.Write(St);
    Sound(880);
    Delay(10);
    NoSound;
    Delay(1000);
    W.Done;
  End; { Popup }

  Procedure Popup4(St1,St2,St3,St4: String);
  Var
    W       : DWindow;
    I,J,X,Y : Integer;

  Begin
    Screen.TextColor(Black);
    Screen.TextBackground(LightGray);
    I := Length(St1);
    If I < Length(St2) Then I := Length(St2);
    If I < Length(St3) Then I := Length(St3);
    If I < Length(St4) Then I := Length(St4);
    X := (MaxX Div 2) - ((I + 4) Div 2);
    Y := MaxY Div 2;
    W.Init(X,Y - 2,X + I + 3,Y + 3,
           Black,LightGray,Black,LightGray,Black,LightGray,
           DoubleBorder,'','',True);
    W.Display;
    Screen.GotoXY(2,1);
    Screen.Write(St1);
    Screen.GotoXY(2,2);
    Screen.Write(St2);
    Screen.GotoXY(2,3);
    Screen.Write(St3);
    Screen.GotoXY(2,4);
    Screen.Write(St4);
    Sound(880);
    Delay(10);
    NoSound;
    Delay(1000);
    W.Done;
  End; { Popup4 }

  Procedure StayPopup(St: String);
  Var I,J,X,Y : Integer;
  Begin
    Screen.TextColor(Black);
    Screen.TextBackground(LightGray);
    I := Length(St);
    X := (MaxX Div 2) - ((I + 4) Div 2);
    Y := MaxY Div 2;
    _Win := New(PWindow,Init(X,Y - 2,X + I + 3,Y + 2,
              Black,LightGray,Black,LightGray,Black,LightGray,
              DoubleBorder,'','',True));
    _Win^.Display;
    Screen.FastWrite(Lo(Screen.WindMin) + 2,Hi(Screen.WindMin) + 2,Screen.TextAttr,St);
  End; { StayPopup }

  Procedure StayPopup4(St1,St2,St3,St4: String);
  Var I,J,X,Y : Integer;
  Begin
    Screen.TextColor(Black);
    Screen.TextBackground(LightGray);
    I := Length(St1);
    If Length(St2) > I Then I := Length(St2);
    If Length(St3) > I Then I := Length(St3);
    If Length(St4) > I Then I := Length(St4);

    X := (MaxX Div 2) - ((I + 4) Div 2);
    Y := MaxY Div 2;
    _Win := New(PWindow,Init(X,Y - 2,X + I + 3,Y + 3,
              Black,LightGray,Black,LightGray,Black,LightGray,
              DoubleBorder,'','',True));
    _Win^.Display;
    Screen.FastWrite(Lo(Screen.WindMin) + 2,Hi(Screen.WindMin) + 1,Screen.TextAttr,St1);
    Screen.FastWrite(Lo(Screen.WindMin) + 2,Hi(Screen.WindMin) + 2,Screen.TextAttr,St2);
    Screen.FastWrite(Lo(Screen.WindMin) + 2,Hi(Screen.WindMin) + 3,Screen.TextAttr,St3);
    Screen.FastWrite(Lo(Screen.WindMin) + 2,Hi(Screen.WindMin) + 4,Screen.TextAttr,St4);
  End; { StayPopup4 }

  Procedure SetScreenMode;
  Var
    I       : Integer;
    Sel     : PSelection;
    MaxX0   : Integer;
    MaxY0   : Integer;
    DX,DY   : Integer;
    OldMode : Word;
    T1      : LongInt;
    T2      : LongInt;
    Ch      : Char;

    Procedure GotoMode(Mode: Word);
    Begin
      MaxX0 := MaxX;
      MaxY0 := MaxY;
      Screen.SetMode(Mode);
      Screen.ShowCursor(False);
      Screen.TextColor(LightGray);
      Screen.TextBackground(Blue);
      Screen.ClrScr;
      Cols   := MaxX Div DisplayColWidth[Config.MenuDisplayScheme];
      Rows   := MaxY - 2;
      OldTop := -1;      { Invalidate to force full redisplay }
      Display;
      Config.ScreenModeWidth  := Screen.Width;
      Config.ScreenModeHeight := Screen.Height;
      Config.ScreenModeDepth  := Screen.BPP;
      SaveConfiguration(Config,Version);
      DX := (MaxX - MaxX0) Div 2;
      DY := (MaxY - MaxY0) Div 2;
      _Menu^.MoveWindow(DX,DY);
      DisMenu^.MoveWindow(DX,DY);
      ColMenu^.MoveWindow(DX,DY);
      PreMenu^.MoveWindow(DX,DY);
      ScnMenu^.MoveWindow(DX,DY);
      EnaMenu^.MoveWindow(DX,DY);
    End; { GotoMode }

  Begin
    Repeat
      OldMode := Screen.CurrentMode;
      I := 0;
      While I < ScnMenu^.SelList^.Count Do
      Begin
        Sel          := PSelection(ScnMenu^.SelList^.At(I));
        Sel^.Checked := (I = Screen.CurrentMode);
        Inc(I);
      End; { While }
      ScnMenu^.Execute;
      If ScnMenu^.CurSel <> 0 Then
      Begin
        StayPopup4('INTVDOS will now resize your desktop.  If the screen does not',
                   'display correctly, wait ten seconds, and the screen will',
                   'revert to this screen mode.',
                   '                                         Hit a key to test');
        ReadKey;
        _Win^.Remove;
        Dispose(_Win,Done);
        GotoMode(ScnMenu^.CurSel - 1);
        T1 := Clock;
        StayPopup4('Did the screen display correctly?','',
                   '         (Y)es  or (N)o','');
        Ch := #0;
        Repeat
          If KeyPressed Then Ch := UpCase(ReadKey);
          T2 := Clock;
        Until (T2 < T1 - 1000) Or (T2 >= T1 + 10000) Or (Ch In ['N','Y']);
        _Win^.Remove;
        Dispose(_Win,Done);
        If (Ch = 'N') Or (T2 >= T1 + 10000) Then GotoMode(OldMode);
      End;
    Until ScnMenu^.CurSel = 0;
  End; { SetScreenMode }

  Procedure SetDisplayScheme;
  Var
    Sel : PSelection;
    I   : Integer;

  Begin
    Repeat
      I := 0;
      While I < DisMenu^.SelList^.Count Do
      Begin
        Sel          := PSelection(DisMenu^.SelList^.At(I));
        Sel^.Checked := (I = Byte(Config.MenuDisplayScheme));
        Inc(I);
      End; { While }
      DisMenu^.Execute;
      If DisMenu^.CurSel > 0 Then
      Begin
        Case DisMenu^.CurSel Of
          1: Config.MenuDisplayScheme := dsNameExt4;
          2: Config.MenuDisplayScheme := dsNameExt3;
          3: Config.MenuDisplayScheme := dsNameExt2;
          4: Config.MenuDisplayScheme := dsNameOnly4;
          5: Config.MenuDisplayScheme := dsNameOnly3;
          6: Config.MenuDisplayScheme := dsNameOnly2;
          7: Config.MenuDisplayScheme := dsRealNameLeft;
          8: Config.MenuDisplayScheme := dsRealNameCenter;
        End; { Case }
        SaveConfiguration(Config,Version);
        Cols   := MaxX Div DisplayColWidth[Config.MenuDisplayScheme];
        OldTop := -1;      { Invalidate to force full redisplay }
        Display;
      End;
    Until DisMenu^.CurSel = 0;
  End; { SetDisplayScheme }

  Procedure SetColorScheme;
  Var
    Sel : PSelection;
    I   : Integer;

    Procedure ChooseColors;
    Var
      Win : PWindow;
      C   : Array[1..4] Of Byte;
      Ch  : Char;
      Col : Byte;
      I   : Byte;

    Begin
      C[1] := Config.MenuForeColor1;
      C[2] := Config.MenuBackColor1;
      C[3] := Config.MenuForeColor2;
      C[4] := Config.MenuBackColor2;
      Col  := 1;
      Win  := New(PWindow,InitWH(50,20,
                 Black,LightGray,Black,LightGray,Black,LightGray,
                 DoubleBorder,' Choose colors ',' Arrow keys to move, <Esc> exits',True));
      Win^.Display;
      Screen.TextColor(Black);
      Screen.TextBackground(LightGray);
      Screen.ClrScr;
      Repeat
        Screen.TextBackground(LightGray);
        For I := 0 To 15 Do
        Begin
          Screen.TextColor(I);
          Screen.GotoXY(10,I + 3); Screen.Write('  ');
          Screen.GotoXY(20,I + 3); Screen.Write('  ');
          Screen.GotoXY(30,I + 3); Screen.Write('  ');
          Screen.GotoXY(40,I + 3); Screen.Write('  ');
        End; { For I }
        For I := 1 To 4 Do
        Begin
          Screen.TextBackground(LightGray);
          If Col = I Then Screen.TextColor(LightRed) Else Screen.TextColor(Black);
          Screen.GotoXY(I * 10 + 2,C[I] + 3); Screen.Write(#17);
        End; { For I }
        Screen.TextColor(Black);
        Screen.GotoXY(1,1);
        Screen.Write('        Fore      Back      Fore      Back');

        Ch := ReadKey;
        If Ch = #0 Then
        Begin
          Ch := ReadKey;
          Case Ch Of
            #71: Begin   { Home }
                   If C[Col] <> 0 Then C[Col] := 0;
                 End;
            #72: Begin   { Up }
                   If C[Col] > 0 Then Dec(C[Col]);
                 End;
            #75: Begin   { Left }
                   If Col > 1 Then Dec(Col);
                 End;
            #77: Begin   { Right }
                   If Col < 4 Then Inc(Col);
                 End;
            #79: Begin   { End }
                   If C[Col] <> 15 Then C[Col] := 15;
                 End;
            #80: Begin   { Down }
                   If C[Col] < 15 Then Inc(C[Col]);
                 End;
          End; { Case }
          Ch := #0;
        End;
      Until Ch = #27;
      Win^.Remove;
      Dispose(Win,Done);
      Config.MenuForeColor1 := C[1];
      Config.MenuBackColor1 := C[2];
      Config.MenuForeColor2 := C[3];
      Config.MenuBackColor2 := C[4];
    End; { ChooseColors }

  Begin
    Repeat
      I := 0;
      While I < ColMenu^.SelList^.Count Do
      Begin
        Sel          := PSelection(ColMenu^.SelList^.At(I));
        Sel^.Checked := (I = Byte(Config.MenuColorScheme));
        Inc(I);
      End; { While }
      ColMenu^.Execute;
      If ColMenu^.CurSel > 0 Then
      Begin
        Case ColMenu^.CurSel Of
          1: Config.MenuColorScheme := csGameType;
          2: Begin
               Config.MenuColorScheme := csAlphabetical;
               ChooseColors;
             End;
          3: Begin
               Config.MenuColorScheme := csEvery3;
               ChooseColors;
             End;
          4: Begin
               Config.MenuColorScheme := csEvery4;
               ChooseColors;
             End;
          5: Begin
               Config.MenuColorScheme := csEvery5;
               ChooseColors;
             End;
          6: Begin
               Config.MenuColorScheme := csEvery6;
               ChooseColors;
             End;
        End; { Case }
        SaveConfiguration(Config,Version);
        OldTop := -1;      { Invalidate to force full redisplay }
        Display;
      End;
    Until ColMenu^.CurSel = 0;
  End; { SetColorScheme }

  Procedure SetFramesPerSecond;
  Var
    _Win : PWindow;
    St   : String;

  Begin
    _Win := New(PWindow,InitWH(60,5,
     Black,LightGray,Black,LightGray,Black,LightGray,
     DoubleBorder,'','',True));
    _Win^.Display;
    Screen.GotoXY(2,2);
    Screen.Write('Enter frames-per-second limit (or 0 for no limit): ');
    Str(Config.FrameRate,St);
    St := _BufLen(53,2,St,Yellow,Blue,3,4);
    Screen.ShowCursor(False);
    Dispose(_Win,Done);
    Val(St,Config.FrameRate,Code);
    If Config.FrameRate < 0 Then Config.FrameRate := 0;
    SaveConfiguration(Config,Version);
    Screen.TextColor(LightGray);
    Screen.TextBackGround(Blue);
  End; { SetFramesPerSecond }

  Procedure SetSoundOptions;
  Var
    _Win : PWindow;
    C    : Char;

  Begin
    _Win := New(PWindow,Init(6,(MaxY Div 2) - 6,MaxX - 5,MaxY Div 2 + 6,
     Black,LightGray,Black,LightGray,Black,LightGray,
     DoubleBorder,'','',True));
    _Win^.Display;
    Screen.GotoXY(2,2);
    Screen.Write('Sound playback rate adjustment');
    Screen.GotoXY(2,3);
    Screen.Write('(The largest value gives the best sound, but might not work on all sound cards)');
    Screen.GotoXY(2,5);
    Screen.Write('Playback rate: [' + IntToStr(Playback_Freq) + ']: ');
    Screen.GotoXY(2,6);
    Screen.Write('  1.............11025 Hz');
    Screen.GotoXY(2,7);
    Screen.Write('  2.............22050 Hz');
    Screen.GotoXY(2,8);
    Screen.Write('  3.............33075 Hz');
    Screen.GotoXY(2,9);
    Screen.Write('  4.............44100 Hz');
    Screen.GotoXY(2,11);
    Screen.Write('Choice (1-4): ');
    Repeat
      C := ReadKey;
    Until C In ['1'..'4'];
    Dispose(_Win,Done);
    Case C Of
      '1': Playback_Freq := 11025;
      '2': Playback_Freq := 22050;
      '3': Playback_Freq := 33075;
      '4': Playback_Freq := 44100;
    End; { Case }
    Config.PlaybackRate := Playback_Freq;
    SaveConfiguration(Config,Version);
    Screen.TextColor(LightGray);
    Screen.TextBackGround(Blue);
  End; { SetSoundOptions }

  Procedure SetExtraLines;
  Begin
    Config.ExtraLines := Not Config.ExtraLines;
    SaveConfiguration(Config,Version);
  End; { SetExtraLines }

  Procedure SetPreferences;
  Var Sel: PSelection;
  Begin
    Repeat
      PreMenu^.GetSelection(7)^.Checked := Config.ExtraLines;
      PreMenu^.Execute;
      Case PreMenu^.CurSel Of
        1: SetScreenMode;
        2: SetColorScheme;
        3: SetDisplayScheme;
        4: SetFramesPerSecond;
        5: SetSoundOptions;
        6: SetExtraLines;
      End; { Case }
    Until PreMenu^.CurSel = 0;
  End; { SetPreferences }

  Procedure InitControllerInfo;
  Begin
    Info.P[1]    := $FF;
    Info.P[2]    := $FF;
    Info.P[3]    := $FF;
    Info.P[4]    := $FF;
    Info.Console := 0;
  End; { InitControllerInfo }

Begin
  OldTop  := -1;
  OldWork := Nil;

  { Try to set to the mode saved in the configuration, otherwise }
  { use mode 0                                                   }

  I     := 0;
  Found := False;
  While (I < Screen.ScreenModes^.Count) And Not Found Do
  Begin
    ScnMode := PScreenMode(Screen.ScreenModes^.At(I));
    If (ScnMode^.Width  = Config.ScreenModeWidth) And
       (ScnMode^.Height = Config.ScreenModeHeight) And
       (ScnMode^.BPP    = Config.ScreenModeDepth)
     Then Found := True Else Inc(I);
  End; { While }
  If Not Found Then I := 0;
  Screen.SetMode(I);
  Screen.ShowCursor(False);
  Screen.TextColor(LightGray);
  Screen.TextBackGround(Blue);
  Screen.ClrScr;

  { Locate all the .BIN files }

  Head := Nil;
  Tail := Nil;
  Num  := 0;
  If ParamCount > 0 Then
  Begin
    Path := ParamStr(1);
    If (Path <> '') And Not (Path[Length(Path)] In [':','\']) Then Path := Path + '\';
  End
  Else Path := '';

  FindFirst(Path + '*.BIN',Archive,S);
  While DosError = 0 Do
  Begin
    If Head = Nil Then
    Begin
      New(Head);
      Tail := Head;
      Tail^.Prev := Nil;
    End
    Else
    Begin
      New(Tail^.Next);
      Tail^.Next^.Prev := Tail;
      Tail := Tail^.Next;
    End;
    Inc(Num);
    Tail^.Name := S.Name;
    Tail^.Size := S.Size;
    Tail^.Next := Nil;
    FindProfile(S.Name,Tail^.Prof);
    FindNext(S);
  End; { While }

  { Sort the list in ascending alphabetical order }

  Work := Head;
  While Work <> Nil Do
  Begin
    Work1 := Work^.Next;
    While Work1 <> Nil Do
    Begin
      If Work1^.Name < Work^.Name Then
      Begin
        Temp.Name   := Work1^.Name;
        Temp.Size   := Work1^.Size;
        Temp.Prof   := Work1^.Prof;
        Work1^.Name := Work^.Name;
        Work1^.Size := Work^.Size;
        Work1^.Prof := Work^.Prof;
        Work^.Name  := Temp.Name;
        Work^.Size  := Temp.Size;
        Work^.Prof  := Temp.Prof;
      End;
      Work1 := Work1^.Next;
    End; { Work1 }
    Work := Work^.Next;
  End; { While }

  { Main loop }

  If Head = Nil Then
  Begin
    Screen.WriteLn('Sorry, but INTVDOS was unable to find any .BIN files in');
    Screen.WriteLn('directory ' + FExpand(Path));
    Screen.WriteLn('');
    Screen.WriteLn('You should make sure that this is the right');
    Screen.WriteLn('path to specify and that there are .BIN files there.');

    Menu := '';
    Exit;
  End;
  FindTop;
  FindSel;
  Cols  := MaxX Div DisplayColWidth[Config.MenuDisplayScheme];
  Rows  := MaxY - 2;
  Show  := True;

  _Menu := New(PMenu,InitWH(31,7,
             Black,LightGray,Black,LightGray,Black,LightGray,
             Red,LightGray,Black,Green,Red,Green,
             DarkGray,LightGray,DarkGray,Green,
             SingleBorder,' INTVDOS ' + Version + ' Main Menu ',
             '',True));
  _Menu^.AddSelection(1, 1,29, 1,False,False,'Configure controllers');
  _Menu^.AddSelection(1, 2,29, 1,False,False,'Enable interfaces');
  _Menu^.AddSelection(1, 3,29, 1,False,False,'Preferences');
  _Menu^.AddSelection(1, 4,29, 8,False,False,'Toggle debugger');
  _Menu^.AddSelection(1, 5,29, 8,False,False,'Toggle sound');

  { Enable interfaces menu }

  EnaMenu := New(PMenu,InitWH(36,InterfaceHandler^.Stack^.Count + 2,
             Black,LightGray,Black,LightGray,Blue,LightGray,
             Red,LightGray,Black,Green,Red,Green,
             DarkGray,LightGray,DarkGray,Green,
             SingleBorder,' Enable interfaces ',' <Enter> toggles, <Esc> exits ',True));
  I := 0;
  While I < InterfaceHandler^.Stack^.Count Do
  Begin
    EnaMenu^.AddSelection(1,I + 1,34,1,False,False,'  ' +
                         PInterface(InterfaceHandler^.Stack^.At(I))^.Name);
    Inc(I);
  End; { While }

  { Preferences menu }

  PreMenu := New(PMenu,InitWH(36,8,
                 Black,LightGray,Black,LightGray,Blue,LightGray,
                 Red,LightGray,Black,Green,Red,Green,
                 DarkGray,LightGray,DarkGray,Green,
                 SingleBorder,' Preferences ',' <Enter> selects, <Esc> exits ',True));
  PreMenu^.AddSelection(1,1,34, 3,False,False,'  Screen mode');
  PreMenu^.AddSelection(1,2,34,13,False,False,'  File list color scheme');
  PreMenu^.AddSelection(1,3,34,13,False,False,'  File list display scheme');
  PreMenu^.AddSelection(1,4,34, 7,False,False,'  Set frames per second');
  PreMenu^.AddSelection(1,5,34,13,False,False,'  Set sound options');
  PreMenu^.AddSelection(1,6,34,13,False,False,'  Use 320x204 screen mode');

  { Screen mode menu }

  ScnMenu := New(PMenu,InitWH(36,Screen.ScreenModes^.Count + 2,
                 Black,LightGray,Black,LightGray,Blue,LightGray,
                 Red,LightGray,Black,Green,Red,Green,
                 DarkGray,LightGray,DarkGray,Green,
                 SingleBorder,' Choose screen mode ',' <Enter> selects, <Esc> exits ',True));
  I := 0;
  While I < Screen.ScreenModes^.Count Do
  Begin
    ScnMode := PScreenMode(Screen.ScreenModes^.At(I));
    St      := '  ' + IntToStr(ScnMode^.Width) + 'x' + IntToStr(ScnMode^.Height);
    Case ScnMode^.DisplayType Of
      dtText: St := St + ' text';
  dtVESAText: St := St + ' VESA text';
   dtVESA256: St := St + 'x256 VESA';
  dtVESATrue: St := St + 'x' + IntToStr(ScnMode^.BPP) + '-bit VESA';
    End; { Case }
    ScnMenu^.AddSelection(1,I + 1,34,1,False,I = Screen.CurrentMode,St);
    Inc(I);
  End; { While }

  { Color scheme menu }

  ColMenu := New(PMenu,InitWH(36,8,
                 Black,LightGray,Black,LightGray,Blue,LightGray,
                 Red,LightGray,Black,Green,Red,Green,
                 DarkGray,LightGray,DarkGray,Green,
                 SingleBorder,' Color schemes ',' <Enter> selects, <Esc> exits ',True));
  ColMenu^.AddSelection(1,1,34, 3,False,False,'  Game type');
  ColMenu^.AddSelection(1,2,34,13,False,False,'  Alternate alphabetically');
  ColMenu^.AddSelection(1,3,34,19,False,False,'  Alternate every 3 rows');
  ColMenu^.AddSelection(1,4,34,19,False,False,'  Alternate every 4 rows');
  ColMenu^.AddSelection(1,5,34,19,False,False,'  Alternate every 5 rows');
  ColMenu^.AddSelection(1,6,34,19,False,False,'  Alternate every 6 rows');

  { Display scheme menu }

  DisMenu := New(PMenu,InitWH(36,10,
                 Black,LightGray,Black,LightGray,Blue,LightGray,
                 Red,LightGray,Black,Green,Red,Green,
                 DarkGray,LightGray,DarkGray,Green,
                 SingleBorder,' Display schemes ',' <Enter> selects, <Esc> exits ',True));
  DisMenu^.AddSelection(1,1,34,12,False,False,'  Name and extension w/4 spaces');
  DisMenu^.AddSelection(1,2,34,12,False,False,'  Name and extension w/3 spaces');
  DisMenu^.AddSelection(1,3,34,12,False,False,'  Name and extension w/2 spaces');
  DisMenu^.AddSelection(1,4,34, 3,False,False,'  Name only w/4 spaces');
  DisMenu^.AddSelection(1,5,34, 3,False,False,'  Name only w/3 spaces');
  DisMenu^.AddSelection(1,6,34, 3,False,False,'  Name only w/2 spaces');
  DisMenu^.AddSelection(1,7,34, 3,False,False,'  Real name (left justified)');
  DisMenu^.AddSelection(1,8,34, 3,False,False,'  Real name (centered)');

  InterfaceHandler^.ActivateInterfaces;
  Repeat
    If Show Then
    Begin
      Display;
      Show := False;
    End;
    Ch := #0;

    { See if the controllers can move the cursor }

    Cap.Cont[1] := imDisc + imKeypad + imSide;
    Cap.Cont[2] := 0;
    Cap.Cont[3] := 0;
    Cap.Cont[4] := 0;
{
    Cap.KB_A    := 0;
    Cap.KB_B    := 0;
    Cap.SY_A    := 0;
    Cap.SY_B    := 0;
}
    { Initialize the info structure }

    InitControllerInfo;

    { We're purposely NOT executing InitKeyboard because we don't want }
    { the keyboard interface responding                                }

    FillChar(Key,SizeOf(Key),#0); { Clear the key array just to be sure }

    If PollIntv2PC Or Not Intv2PC Then
    Begin
      InterfaceHandler^.GetControllerInfo(Cap,Info);
      ContDir := Info.P[1] Xor $FF;
      If (ContDir And $1F) In [Disc_NNW,Disc_N,Disc_NNE,Disc_NE] Then
      Begin
        If Not Intv2PC Then Delay(1);
        InitControllerInfo;
        Repeat
        Until PollIntv2PC Or Not Intv2PC;
        InterfaceHandler^.GetControllerInfo(Cap,Info);
        ContDir := Info.P[1] Xor $FF;
        If (ContDir And $1F) In [Disc_NNW,Disc_N,Disc_NNE,Disc_NE] Then
        Begin
          _Up;
          If Intv2PC Then Delay(10) Else Delay(30);
        End;
      End;
      If (ContDir And $1F) In [Disc_SSE,Disc_S,Disc_SSW,Disc_SW] Then
      Begin
        If Not Intv2PC Then Delay(1);
        InitControllerInfo;
        Repeat
        Until PollIntv2PC Or Not Intv2PC;
        InterfaceHandler^.GetControllerInfo(Cap,Info);
        ContDir := Info.P[1] Xor $FF;
        If (ContDir And $1F) In [Disc_SSE,Disc_S,Disc_SSW,Disc_SW] Then
        Begin
          _Down;
          If Intv2PC Then Delay(10) Else Delay(30);
        End;
      End;
      If (ContDir And $1F) In [Disc_WSW,Disc_W,Disc_WNW,Disc_NW] Then
      Begin
        If Not Intv2PC Then Delay(1);
        InitControllerInfo;
        Repeat
        Until PollIntv2PC Or Not Intv2PC;
        InterfaceHandler^.GetControllerInfo(Cap,Info);
        ContDir := Info.P[1] Xor $FF;
        If (ContDir And $1F) In [Disc_WSW,Disc_W,Disc_WNW,Disc_NW] Then
        Begin
          _Left;
          If Intv2PC Then Delay(10) Else Delay(30);
        End;
      End;
      If (ContDir And $1F) In [Disc_ENE,Disc_E,Disc_ESE,Disc_SE] Then
      Begin
        If Not Intv2PC Then Delay(1);
        InitControllerInfo;
        Repeat
        Until PollIntv2PC Or Not Intv2PC;
        InterfaceHandler^.GetControllerInfo(Cap,Info);
        ContDir := Info.P[1] Xor $FF;
        If (ContDir And $1F) In [Disc_ENE,Disc_E,Disc_ESE,Disc_SE] Then
        Begin
          _Right;
          If Intv2PC Then Delay(10) Else Delay(30);
        End;
      End;

      If (ContDir And $E0) In [Side_Up,Side_BL,Side_BR] Then
      Begin
        If Not Intv2PC Then Delay(1);
        InitControllerInfo;
        Repeat
        Until PollIntv2PC Or Not Intv2PC;
        InterfaceHandler^.GetControllerInfo(Cap,Info);
        ContDir := Info.P[1] Xor $FF;
        If (ContDir And $E0) In [Side_Up,Side_BL,Side_BR] Then
        Begin
          Ch := #13;

          { Wait until button is released }

          Repeat
            Cap.Cont[1] := imDisc + imKeypad + imSide;
            Cap.Cont[2] := 0;
            Cap.Cont[3] := 0;
            Cap.Cont[4] := 0;
  {
            Cap.KB_A    := 0;
            Cap.KB_B    := 0;
            Cap.SY_A    := 0;
            Cap.SY_B    := 0;
  }
            { Initialize the info structure }

            InitControllerInfo;
            Repeat
            Until PollIntv2PC Or Not Intv2PC;
            InterfaceHandler^.GetControllerInfo(Cap,Info);
            ContDir := Info.P[1] Xor $FF;
          Until Not ((ContDir And $E0) In [Side_Up,Side_BL,Side_BR]);
        End;
      End;
    End;

    If KeyPressed Then
    Begin
      Ch := UpCase(ReadKey);
      If Ch = #0 Then
      Begin
        C := ReadKey;
        Case C Of
(*
          #59: Begin                     { F1 }
                 MainMenuHelp;
                 OldTop := -1;  { Invalidate to force full redisplay }
                 Show := True;
               End;
*)
          #68: Begin                     { F10 }
                 Repeat
                   _Menu^.Execute;
                   Case  _Menu^.CurSel Of
                     1: Begin
                          InterfaceHandler^.DeactivateInterfaces;
                          InterfaceHandler^.Configure;
                          SaveConfiguration(Config,Version);
                          InterfaceHandler^.ActivateInterfaces;
                        End;
                     2: Begin
                          If InterfaceHandler^.Stack^.Count > 0 Then
                          Begin

                            { Check/uncheck various interfaces as appropriate }

                            I := 0;
                            While I < InterfaceHandler^.Stack^.Count Do
                            Begin
                              EnaMenu^.GetSelection(I + 1)^.Checked :=
                               PInterface(InterfaceHandler^.Stack^.At(I))^.Enabled;
                              Inc(I);
                            End; { While }
                            Repeat
                              EnaMenu^.Execute;
                              If EnaMenu^.CurSel > 0 Then
                              Begin
                                InterfaceHandler^.DeactivateInterfaces;
                                Intf := PInterface(InterfaceHandler^.Stack^.At(EnaMenu^.CurSel - 1));
                                Intf^.Enabled := Not Intf^.Enabled;
                                EnaMenu^.GetSelection(EnaMenu^.CurSel)^.Checked := Intf^.Enabled;
                                InterfaceHandler^.ActivateInterfaces;
                              End;
                            Until EnaMenu^.CurSel = 0;
                            SaveConfiguration(Config,Version);
                          End;
                        End;
                     3: SetPreferences;
                     4: Begin
                          Debugger := Not Debugger;
                          If Debugger
                           Then Popup('Debugger is ON')
                           Else Popup('Debugger is OFF');
                        End;
                     5: Begin
                          DoSound := Not DoSound;
                          Config.DoSound := DoSound;
                          If DoSound
                           Then Popup('Sound is ON')
                           Else Popup('Sound is OFF');
                          SaveConfiguration(Config,Version);
                        End;
                   End; { Case }
                 Until _Menu^.CurSel = 0;
               End;
          #71: Begin                     { Home }
                 If Sel <> 0 Then Show := True;
                 Top   := 0;
                 Sel   := 0;
                 Work1 := Head;
                 Work  := Head;
               End;
          #72: _Up;                      { Up }
          #73: Begin                     { PgUp }
                 If Sel >= Cols * Rows Then
                 Begin
                   Screen.TextColor(LightGray);
                   Screen.TextBackground(Blue);
                   Screen.ClrScr;
                   Show := True;
                   Dec(Sel,Cols * Rows);
                   Dec(Top,Cols * Rows);
                   If Top < 0 Then Top := 0;
                   FindTop;
                   FindSel;
                 End;
               End;
          #75: _Left;                    { Left }
          #77: _Right;                   { Right }
          #79: Begin                     { End }
                 If Sel <> Num - 1 Then Show := True;
                 Sel := Num - 1;
                 I   := Top;
                 Top := Sel - Cols * Rows + 1;
                 If Top < 0 Then Top := 0;
                 If I <> Top Then Show := True;
                 Work  := Tail;
                 FindTop;
               End;
          #80: _Down;                    { Down }
          #81: Begin                     { PgDn }
                 If Sel < Num - Cols * Rows Then
                 Begin
                   Screen.TextColor(LightGray);
                   Screen.TextBackground(Blue);
                   Screen.ClrScr;
                   Show := True;
                   Inc(Sel,Cols * Rows);
                   Inc(Top,Cols * Rows);
                   FindTop;
                   FindSel;
                 End;
               End;
        End; { Case }
      End;
    End;
  Until Ch In [#13,#27,{'4',}'E'];
  If Ch In [#13,'4','E'] Then
  Begin
    UseMenu    := True;
    Menu       := Path + Work^.Name;
    ProfileRec := Nil;
    Case Ch Of
{      '4': Cart := Cart4;}
      #13: Cart := FindProfile(Work^.Name,ProfileRec);{Cart5;}
      'E': Cart := CartECS;
    End; { Case }
  End
  Else
  Begin
    Menu    := '';
    UseMenu := False;
  End;

  { Kill the file list }

  Work := Head;
  While Work <> Nil Do
  Begin
    Work1 := Work^.Next;
    Dispose(Work);
    Work := Work1;
  End; { While }
  Screen.TextColor(LightGray);
  Screen.TextBackground(Black);
  Screen.ClrScr;
  Dispose(_Menu,Done);
  Dispose(PreMenu,Done);
  Dispose(DisMenu,Done);
  Dispose(ColMenu,Done);
  Dispose(ScnMenu,Done);
  Dispose(EnaMenu,Done);
End; { Menu }

Begin
  FHead := Nil;
  FTail := Nil;
  New(FileIndex);
  If Not Exist(OutFileName) Then MakeProfile('.\');
  LoadProfile;
  IndexFileList;
  OldExitProc := ExitProc;
  ExitProc    := @KillProfile;
End.